import contextlib
import os
import pathlib
import sys
import threading
from collections import defaultdict
from collections import deque
from dataclasses import dataclass
from datetime import datetime
from functools import total_ordering
from math import ceil
from typing import Any
from typing import DefaultDict
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import cast

from rich.markup import escape
from rich.segment import Segment
from rich.style import Style
from rich.text import Text
from textual import events
from textual import log
from textual.app import App
from textual.app import ComposeResult
from textual.binding import Binding
from textual.color import Color
from textual.color import Gradient
from textual.containers import Container
from textual.containers import HorizontalScroll
from textual.dom import DOMNode
from textual.message import Message
from textual.reactive import reactive
from textual.screen import Screen
from textual.strip import Strip
from textual.widget import Widget
from textual.widgets import DataTable
from textual.widgets import Footer
from textual.widgets import Label
from textual.widgets import Static
from textual.widgets.data_table import RowKey

from memray import AllocationRecord
from memray import SocketReader
from memray._memray import size_fmt
from memray.reporters._textual_hacks import Bindings
from memray.reporters._textual_hacks import redraw_footer
from memray.reporters._textual_hacks import update_key_description

MAX_MEMORY_RATIO = 0.95


@dataclass(frozen=True)
class Location:
    function: str
    file: str


@dataclass
class AllocationEntry:
    own_memory: int
    total_memory: int
    n_allocations: int
    thread_ids: Set[int]


@dataclass(frozen=True, eq=False)
class Snapshot:
    heap_size: int
    records: List[AllocationRecord]
    records_by_location: Dict[Location, AllocationEntry]


_EMPTY_SNAPSHOT = Snapshot(heap_size=0, records=[], records_by_location={})


class SnapshotFetched(Message):
    def __init__(self, snapshot: Snapshot, disconnected: bool) -> None:
        self.snapshot = snapshot
        self.disconnected = disconnected
        super().__init__()


class MemoryGraph(Widget):
    def __init__(
        self,
        *args: Any,
        max_data_points: int,
        height: int = 4,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        maxval: float = 1.0
        minval: float = 0.0
        self._width = max_data_points
        self._height = height
        self._minval = minval
        self._maxval = maxval
        values = [minval] * (2 * self._width + 1)
        self._values = deque(values, maxlen=2 * self._width)

        self._lookup = [
            [" ", "▗", "▐"],
            ["▖", "▄", "▟"],
            ["▌", "▙", "█"],
        ]

        self.border_title = "Heap Usage"

    def _value_to_blocks(self, value: float) -> List[int]:
        dots_per_block = 2
        if value < self._minval:
            n_dots = 0
        elif value > self._maxval:
            n_dots = dots_per_block * self._height
        else:
            n_dots = ceil(
                (value - self._minval)
                / (self._maxval - self._minval)
                * dots_per_block
                * self._height
            )
        blocks = [dots_per_block] * (n_dots // dots_per_block)
        if n_dots % dots_per_block > 0:
            blocks += [n_dots % dots_per_block]
        blocks += [0] * (self._height - len(blocks))
        return blocks

    def add_value(self, value: float) -> None:
        if value > self._maxval:
            self._maxval = value
        self._values.append(value)
        if self._maxval > 1:
            self.border_subtitle = (
                f"{size_fmt(int(value))}"
                f" ({int(round(value * 100/self._maxval, 0))}%"
                f" of {size_fmt(int(self._maxval))} max)"
            )
        self.refresh()

    def render_line(self, y: int) -> Strip:
        graph: list[list[str]] = [[] for _ in range(self._height)]
        blocks_by_index = [self._value_to_blocks(value) for value in self._values]

        for left, right in zip(blocks_by_index[::2], blocks_by_index[1::2]):
            for row, char in enumerate(
                reversed(tuple(self._lookup[li][ri] for li, ri in zip(left, right)))
            ):
                graph[row].append(char)

        if y > len(graph):
            return Strip.blank(self.size.width)
        data = " " * self.size.width
        data += "".join(graph[y])
        data = data[-self.size.width :]
        return Strip([Segment(data, self.rich_style)])


@total_ordering
class SortableText(Text):
    __slots__ = ("value",)

    def __init__(
        self,
        value: Any,
        text: str,
        color: Color,
        justify: Any = "right",  # "Any" is a hack: justify should be Literal
    ) -> None:
        self.value = value
        super().__init__(
            str(text),
            Style(color=color.rich_color),
            justify=justify,
        )

    def __lt__(self, other: Any) -> bool:
        if type(other) != SortableText:
            return NotImplemented
        return cast(bool, self.value < other.value)

    def __gt__(self, other: Any) -> bool:
        if type(other) != SortableText:
            return NotImplemented
        return cast(bool, self.value > other.value)

    def __eq__(self, other: Any) -> bool:
        if type(other) != SortableText:
            return NotImplemented
        return cast(bool, self.value == other.value)


def aggregate_allocations(
    allocations: Iterable[AllocationRecord],
    memory_threshold: float = float("inf"),
    native_traces: Optional[bool] = False,
) -> Dict[Location, AllocationEntry]:
    """Take allocation records and for each frame contained, record "own"
    allocations which happened on the frame, and sum up allocations on
    all of the child frames to calculate "total" allocations."""

    processed_allocations: DefaultDict[Location, AllocationEntry] = defaultdict(
        lambda: AllocationEntry(
            own_memory=0, total_memory=0, n_allocations=0, thread_ids=set()
        )
    )

    current_total = 0
    for allocation in allocations:
        if current_total >= memory_threshold:
            break
        current_total += allocation.size

        stack_trace = list(
            allocation.hybrid_stack_trace()
            if native_traces
            else allocation.stack_trace()
        )
        if not stack_trace:
            frame = processed_allocations[Location(function="???", file="???")]
            frame.total_memory += allocation.size
            frame.own_memory += allocation.size
            frame.n_allocations += allocation.n_allocations
            frame.thread_ids.add(allocation.tid)
            continue

        # Walk upwards and sum totals
        visited = set()
        for i, (function, file_name, _) in enumerate(stack_trace):
            location = Location(function=function, file=file_name)
            frame = processed_allocations[location]
            if location in visited:
                continue
            visited.add(location)
            if i == 0:
                frame.own_memory += allocation.size
            frame.total_memory += allocation.size
            frame.n_allocations += allocation.n_allocations
            frame.thread_ids.add(allocation.tid)
    return processed_allocations


class TimeDisplay(Static):
    """TUI widget to display the current time."""

    def on_mount(self) -> None:
        """Event handler called when the widget is added to the app."""
        self.set_interval(0.1, lambda: self.update(datetime.now().ctime()))


def _filename_to_module_name(file: str) -> str:
    if file.endswith(".py"):
        for path in sys.path:
            if not os.path.isdir(path):
                continue
            with contextlib.suppress(ValueError):
                relative_path = pathlib.Path(file).relative_to(path)
                ret = str(relative_path.with_suffix(""))
                ret = ret.replace(os.sep, ".").replace(".__init__", "")
                return ret
    return file


class AllocationTable(Widget):
    """Widget to display the TUI table."""

    COMPONENT_CLASSES = {
        "allocationtable--sorted-column-heading",
        "allocationtable--function-name",
    }

    DEFAULT_CSS = """
    AllocationTable .allocationtable--sorted-column-heading {
        text-style: bold underline;
    }

    AllocationTable .allocationtable--function-name {
    }
    """

    default_sort_column_id = 1
    sort_column_id = reactive(default_sort_column_id)
    snapshot = reactive(_EMPTY_SNAPSHOT)
    current_thread = reactive(0)
    merge_threads = reactive(False, init=False)

    columns = [
        "Location",
        "Total Bytes",
        "% Total",
        "Own Bytes",
        "% Own",
        "Allocations",
        "File/Module",
    ]

    KEY_TO_COLUMN_NAME = {
        1: "total_memory",
        3: "own_memory",
        5: "n_allocations",
    }

    HIGHLIGHTED_COLUMNS_BY_SORT_COLUMN = {
        1: (1, 2),
        3: (3, 4),
        5: (5,),
    }

    SORT_COLUMN_BY_CLICKED_COLUMN = {
        clicked_col: sort_col
        for sort_col, clicked_cols in HIGHLIGHTED_COLUMNS_BY_SORT_COLUMN.items()
        for clicked_col in clicked_cols
    }

    def __init__(self) -> None:
        super().__init__()
        self._composed = False

        gradient = Gradient(
            (0, Color(97, 193, 44)),
            (0.4, Color(236, 152, 16)),
            (0.6, Color.parse("darkorange")),
            (1, Color.parse("indianred")),
        )
        self._color_by_percentage = {i: gradient.get_color(i / 100) for i in range(101)}

    def _get_color(self, value: float, max: float) -> Color:
        return self._color_by_percentage[int(value * 100 / max)]

    def get_heading(self, column_idx: int) -> Text:
        sort_column = (
            self.sort_column_id if self._composed else self.default_sort_column_id
        )
        sort_column_style = self.get_component_rich_style(
            "allocationtable--sorted-column-heading",
            partial=True,
        )
        log(
            f"self._composed={self._composed} sort_column={sort_column}"
            f" highlighted_cols={self.HIGHLIGHTED_COLUMNS_BY_SORT_COLUMN[sort_column]}"
        )
        if column_idx in (0, len(self.columns) - 1):
            return Text(self.columns[column_idx], justify="center")
        elif column_idx in self.HIGHLIGHTED_COLUMNS_BY_SORT_COLUMN[sort_column]:
            return Text(
                self.columns[column_idx], justify="right", style=sort_column_style
            )
        else:
            return Text(self.columns[column_idx], justify="right").on(
                click=f"screen.sort({self.SORT_COLUMN_BY_CLICKED_COLUMN[column_idx]})"
            )

    def compose(self) -> ComposeResult:
        table: DataTable[Text] = DataTable(
            id="body_table", header_height=1, show_cursor=False, zebra_stripes=True
        )
        table.focus()
        for column_idx in range(len(self.columns)):
            table.add_column(self.get_heading(column_idx), key=str(column_idx))

        # Set an initial size for the Location column to avoid too many resizes
        table.ordered_columns[0].content_width = 50

        self._composed = True
        yield table

    def watch_current_thread(self) -> None:
        """Called when the current_thread attribute changes."""
        self.populate_table()

    def watch_merge_threads(self) -> None:
        """Called when the merge_threads attribute changes."""
        self.populate_table()

    def watch_snapshot(self) -> None:
        """Called when the snapshot attribute changes."""
        self.populate_table()

    def watch_sort_column_id(self, sort_column_id: int) -> None:
        """Called when the sort_column_id attribute changes."""
        table = self.query_one("#body_table", DataTable)

        for i in range(1, len(self.columns)):
            table.ordered_columns[i].label = self.get_heading(i)

        table.sort(table.ordered_columns[sort_column_id].key, reverse=True)

    def populate_table(self) -> None:
        """Method to render the TUI table."""
        table = self.query_one("#body_table", DataTable)

        if not table.columns:
            return

        allocation_entries = self.snapshot.records_by_location
        total_allocations = self.snapshot.heap_size
        num_allocations = sum(
            entry.n_allocations for entry in allocation_entries.values()
        )
        sorted_allocations = sorted(
            allocation_entries.items(),
            key=lambda item: getattr(
                item[1], self.KEY_TO_COLUMN_NAME[self.sort_column_id]
            ),
            reverse=True,
        )

        function_column_style = self.get_component_rich_style(
            "allocationtable--function-name", partial=True
        )

        # Clear previous table rows
        old_locations = set(table.rows)
        new_locations = set()

        for location, result in sorted_allocations:
            if not self.merge_threads and (
                self.current_thread not in result.thread_ids
            ):
                continue

            total_color = self._get_color(result.total_memory, total_allocations)
            own_color = self._get_color(result.own_memory, total_allocations)
            allocation_color = self._get_color(result.n_allocations, num_allocations)

            percent_total = result.total_memory / total_allocations * 100
            percent_own = result.own_memory / total_allocations * 100

            cells = [
                SortableText(
                    result.total_memory, size_fmt(result.total_memory), total_color
                ),
                SortableText(result.total_memory, f"{percent_total:.2f}%", total_color),
                SortableText(result.own_memory, size_fmt(result.own_memory), own_color),
                SortableText(result.own_memory, f"{percent_own:.2f}%", own_color),
                SortableText(
                    result.n_allocations, str(result.n_allocations), allocation_color
                ),
            ]

            row_key = str((location.function, location.file))
            new_locations.add(RowKey(row_key))

            if row_key not in table.rows:
                table.add_row(
                    Text(location.function, style=function_column_style),
                    *cells,
                    Text(_filename_to_module_name(location.file)),
                    key=row_key,
                )
            else:
                for col_idx, val in enumerate(cells, 1):
                    col_key = str(col_idx)
                    table.update_cell(row_key, col_key, val)

        for old_row_key in old_locations - new_locations:
            table.remove_row(old_row_key)

        table.sort(str(self.sort_column_id), reverse=True)


class Header(Widget):
    """Widget to display TUI header information."""

    pid = reactive("")
    command_line = reactive("")
    n_samples = reactive(0)
    start = datetime.now()
    last_update = reactive(start)

    def __init__(self, pid: Optional[int], cmd_line: Optional[str]):
        super().__init__()
        self.pid = str(pid) if pid is not None else "???"
        if not cmd_line:
            cmd_line = "???"
        self.command_line = escape(cmd_line)

    def compose(self) -> ComposeResult:
        header_metadata = HorizontalScroll(
            Container(
                Label(f"[b]PID[/]: {self.pid}", id="pid"),
                Label(f"[b]CMD[/]: {self.command_line}", shrink=False, id="cmd"),
                Label(id="tid"),
                Label(id="thread"),
                Label(id="samples"),
                Label(id="duration"),
                id="header_metadata_grid",
            ),
            Label(id="status_message"),
            id="header_metadata",
        )
        header_metadata.border_title = "(∩｀-´)⊃━☆ﾟ.*･｡ﾟ"
        yield Container(
            header_metadata,
            Container(MemoryGraph(max_data_points=50), id="memory_graph_container"),
            id="header_container",
        )

    def watch_n_samples(self, n_samples: int) -> None:
        """Called when the n_samples attribute changes."""
        self.query_one("#samples", Label).update(f"[b]Samples[/]: {n_samples}")

    def watch_last_update(self, last_update: datetime) -> None:
        """Called when the last_update attribute changes."""
        self.query_one("#duration", Label).update(
            f"[b]Duration[/]: {(last_update - self.start).total_seconds():.1f} seconds"
        )


class TUI(Screen[None]):
    """TUI main application class."""

    CSS_PATH = "tui.css"

    BINDINGS = [
        Binding("ctrl+z", "app.suspend_process"),
        Binding("q,esc", "app.quit", "Quit"),
        Binding("m", "toggle_merge_threads", "Merge Threads"),
        Binding("<,left", "previous_thread", "Previous Thread"),
        Binding(">,right", "next_thread", "Next Thread"),
        Binding("t", "sort(1)", "Sort by Total"),
        Binding("o", "sort(3)", "Sort by Own"),
        Binding("a", "sort(5)", "Sort by Allocations"),
        Binding("space", "toggle_pause", "Pause"),
        Binding("up", "scroll_grid('up')"),
        Binding("down", "scroll_grid('down')"),
    ]

    # Start with a non-empty list of threads so that we always have something
    # to display. This avoids "Thread 1 of 0" and fixes a DivideByZeroError
    # when switching threads before the first allocation is seen.
    _DUMMY_THREAD_LIST = [0]

    thread_idx = reactive(0)
    threads = reactive(_DUMMY_THREAD_LIST, always_update=True)
    snapshot = reactive(_EMPTY_SNAPSHOT)
    paused = reactive(False, init=False)
    disconnected = reactive(False, init=False)

    def __init__(self, pid: Optional[int], cmd_line: Optional[str], native: bool):
        self.pid = pid
        self.cmd_line = cmd_line
        self.native = native
        self._name_by_tid: Dict[int, str] = {}
        self._max_memory_seen = 0
        self._merge_threads = True
        super().__init__()

    @property
    def current_thread(self) -> int:
        return self.threads[self.thread_idx]

    def action_previous_thread(self) -> None:
        """An action to switch to previous thread."""
        if not self._merge_threads:
            self.thread_idx = (self.thread_idx - 1) % len(self.threads)

    def action_next_thread(self) -> None:
        """An action to switch to next thread."""
        if not self._merge_threads:
            self.thread_idx = (self.thread_idx + 1) % len(self.threads)

    def action_sort(self, col_number: int) -> None:
        """An action to sort the table rows based on a given column attribute."""
        self.update_sort_key(col_number)

    def _populate_header_thread_labels(self, thread_idx: int) -> None:
        if self._merge_threads:
            tid_label = "[b]TID[/]: *"
            thread_label = "[b]All threads[/]"
        else:
            tid_label = f"[b]TID[/]: {hex(self.current_thread)}"
            thread_label = f"[b]Thread[/] {thread_idx + 1} of {len(self.threads)}"
            thread_name = self._name_by_tid.get(self.current_thread)
            if thread_name:
                thread_label += f" ({thread_name})"

        self.query_one("#tid", Label).update(tid_label)
        self.query_one("#thread", Label).update(thread_label)

    def action_toggle_merge_threads(self) -> None:
        """An action to toggle showing allocations from all threads together."""
        self._merge_threads = not self._merge_threads
        redraw_footer(self.app)
        self.app.screen.query_one(AllocationTable).merge_threads = self._merge_threads
        self._populate_header_thread_labels(self.thread_idx)

    def action_toggle_pause(self) -> None:
        """Toggle pause on keypress"""
        if self.paused or not self.disconnected:
            self.paused = not self.paused
            redraw_footer(self.app)
            if not self.paused:
                self.display_snapshot()

    def action_scroll_grid(self, direction: str) -> None:
        """Toggle pause on keypress"""
        grid = self.query_one(DataTable)
        getattr(grid, f"action_scroll_{direction}")()

    def watch_thread_idx(self, thread_idx: int) -> None:
        """Called when the thread_idx attribute changes."""
        self._populate_header_thread_labels(thread_idx)
        self.query_one(AllocationTable).current_thread = self.current_thread

    def watch_threads(self) -> None:
        """Called when the threads attribute changes."""
        self._populate_header_thread_labels(self.thread_idx)

    def watch_disconnected(self) -> None:
        self.update_label()
        redraw_footer(self.app)

    def watch_paused(self) -> None:
        self.update_label()

    def watch_snapshot(self, snapshot: Snapshot) -> None:
        """Called automatically when the snapshot attribute is updated"""
        self._latest_snapshot = snapshot
        self.display_snapshot()

    def update_label(self) -> None:
        status_message = []
        if self.paused:
            status_message.append("[yellow]Table updates paused[/]")
        if self.disconnected:
            status_message.append("[red]Remote has disconnected[/]")
        if status_message:
            status_message.insert(0, "[b]Status[/]:")

        log(f"updating status message to {' '.join(status_message)}")
        self.query_one("#status_message", Label).update(" ".join(status_message))

    def compose(self) -> ComposeResult:
        yield Container(
            Label("[b]Memray[/b] live tracking", id="head_title"),
            TimeDisplay(id="head_time_display"),
            id="head",
        )
        yield Header(pid=self.pid, cmd_line=escape(self.cmd_line or ""))
        yield AllocationTable()
        yield Footer()

    def display_snapshot(self) -> None:
        snapshot = self._latest_snapshot

        if snapshot is _EMPTY_SNAPSHOT:
            return

        header = self.query_one(Header)
        body = self.query_one(AllocationTable)
        graph = self.query_one(MemoryGraph)

        # We want to update many header fields even when paused
        header.n_samples += 1
        header.last_update = datetime.now()

        graph.add_value(snapshot.heap_size)

        # Other fields should only be updated when not paused.
        if self.paused:
            return

        name_by_tid = {record.tid: record.thread_name for record in snapshot.records}
        new_tids = name_by_tid.keys() - self._name_by_tid.keys()
        self._name_by_tid.update(name_by_tid)

        if new_tids:
            threads = self.threads
            if threads is self._DUMMY_THREAD_LIST:
                threads = []
            for tid in sorted(new_tids):
                threads.append(tid)
            self.threads = threads

        body.current_thread = self.current_thread
        if not self.paused:
            body.snapshot = snapshot

    def update_sort_key(self, col_number: int) -> None:
        """Method called to update the table sort key attribute."""
        body = self.query_one(AllocationTable)
        body.sort_column_id = col_number

    def rewrite_bindings(self, bindings: Bindings) -> None:
        if "space" in bindings and bindings["space"][1].description == "Pause":
            if self.paused:
                update_key_description(bindings, "space", "Unpause")
            elif self.disconnected:
                del bindings["space"]

        if self._merge_threads:
            bindings.pop("less_than_sign")
            bindings.pop("greater_than_sign")
            update_key_description(bindings, "m", "Unmerge Threads")

    @property
    def active_bindings(self) -> Dict[str, Any]:
        bindings = super().active_bindings.copy()
        self.rewrite_bindings(bindings)
        return bindings


class UpdateThread(threading.Thread):
    def __init__(self, app: "TUIApp", reader: SocketReader) -> None:
        self._app = app
        self._reader = reader
        self._update_requested = threading.Event()
        self._update_requested.set()
        self._canceled = threading.Event()
        super().__init__()

    def run(self) -> None:
        while self._update_requested.wait():
            if self._canceled.is_set():
                return
            self._update_requested.clear()

            records = list(self._reader.get_current_snapshot(merge_threads=False))
            heap_size = sum(record.size for record in records)
            records_by_location = aggregate_allocations(
                records, MAX_MEMORY_RATIO * heap_size, self._reader.has_native_traces
            )
            snapshot = Snapshot(
                heap_size=heap_size,
                records=records,
                records_by_location=records_by_location,
            )

            self._app.post_message(
                SnapshotFetched(
                    snapshot,
                    not self._reader.is_active,
                )
            )

            if not self._reader.is_active:
                return

    def cancel(self) -> None:
        self._canceled.set()
        self._update_requested.set()

    def schedule_update(self) -> None:
        self._update_requested.set()


class TUIApp(App[None]):
    """TUI main application class."""

    CSS_PATH = "tui.css"

    def __init__(
        self,
        reader: SocketReader,
        cmdline_override: Optional[str] = None,
        poll_interval: float = 1.0,
    ) -> None:
        self._reader = reader
        self._poll_interval = poll_interval
        self._cmdline_override = cmdline_override
        self._update_thread = UpdateThread(self, self._reader)
        self.tui: Optional[TUI] = None
        super().__init__()

    def on_mount(self) -> None:
        self.set_interval(self._poll_interval, self._update_thread.schedule_update)
        cmd_line = (
            self._cmdline_override
            if self._cmdline_override is not None
            else self._reader.command_line
        )

        if cmd_line is not None and "/memray" in cmd_line:
            cmd_args = cmd_line.split()
            if any(cmd_args[0].endswith(p) for p in ("/memray", "/memray/__main__.py")):
                cmd_args[0] = "memray"
                cmd_line = " ".join(cmd_args)

        self.tui = TUI(
            pid=self._reader.pid,
            cmd_line=cmd_line,
            native=self._reader.has_native_traces,
        )
        self.push_screen(self.tui)
        self._update_thread.start()

    def on_unmount(self) -> None:
        self._update_thread.cancel()
        if self._update_thread.is_alive():
            self._update_thread.join()

    def on_snapshot_fetched(self, message: SnapshotFetched) -> None:
        """Method called to process each fetched snapshot."""
        assert self.tui is not None
        with self.batch_update():
            self.tui.snapshot = message.snapshot
        if message.disconnected:
            self.tui.disconnected = True

    def on_resize(self, event: events.Resize) -> None:
        self.set_class(0 <= event.size.width < 81, "narrow")

    if hasattr(App, "namespace_bindings"):
        # Removed in Textual 0.61
        @property
        def namespace_bindings(self) -> Dict[str, Tuple[DOMNode, Binding]]:
            bindings = super().namespace_bindings.copy()  # type: ignore[misc]
            if self.tui:
                self.tui.rewrite_bindings(bindings)
            return bindings  # type: ignore[no-any-return]



"""Templates to render reports in HTML."""
from functools import lru_cache
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Union

import jinja2
from markupsafe import Markup

from memray import MemorySnapshot
from memray import Metadata


@lru_cache(maxsize=1)
def get_render_environment() -> jinja2.Environment:
    loader = jinja2.PackageLoader("memray.reporters")
    env = jinja2.Environment(loader=loader)

    def include_file(name: str) -> Markup:
        """Include a file from the templates directory without
        interpolating its contents"""
        source, *_ = loader.get_source(env, name)
        return Markup(source)

    env.globals["include_file"] = include_file
    env.policies["json.dumps_kwargs"] = {"sort_keys": True, "separators": (",", ":")}
    return env


def get_report_title(
    *, kind: str, show_memory_leaks: bool, inverted: bool = False
) -> str:
    parts = []
    if inverted:
        parts.append("inverted")
    parts.append(kind)
    parts.append("report")
    if show_memory_leaks:
        parts.append("(memory leaks)")
    return " ".join(parts)


def render_report(
    *,
    kind: str,
    data: Union[Dict[str, Any], Iterable[Dict[str, Any]]],
    metadata: Metadata,
    memory_records: Iterable[MemorySnapshot],
    show_memory_leaks: bool,
    merge_threads: bool,
    inverted: bool,
    no_web: bool = False,
) -> str:
    env = get_render_environment()
    template = env.get_template(kind + ".html")

    pretty_kind = kind.replace("_", " ")
    title = get_report_title(
        kind=pretty_kind,
        show_memory_leaks=show_memory_leaks,
        inverted=inverted,
    )
    return template.render(
        kind=pretty_kind,
        title=title,
        data=data,
        metadata=metadata,
        memory_records=memory_records,
        show_memory_leaks=show_memory_leaks,
        merge_threads=merge_threads,
        inverted=inverted,
        no_web=no_web,
    )



import argparse
import os
from pathlib import Path
from textwrap import dedent

from memray import FileReader
from memray._errors import MemrayCommandError
from memray.commands.common import warn_if_file_is_not_aggregated_and_is_too_big
from memray.commands.common import warn_if_not_enough_symbols
from memray.reporters.tree import TreeReporter


class TreeCommand:
    """Generate a tree view in the terminal for peak memory usage"""

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("results", help="Results of the tracker run")
        parser.add_argument(
            "-b",
            "--biggest-allocs",
            help="Show n biggest allocations (defaults to 200)",
            type=int,
            default=200,
        )
        alloc_type_group = parser.add_mutually_exclusive_group()
        alloc_type_group.add_argument(
            "--temporary-allocation-threshold",
            metavar="N",
            help=dedent(
                """
                Report temporary allocations, as opposed to leaked allocations
                or high watermark allocations.  An allocation is considered
                temporary if at most N other allocations occur before it is
                deallocated.  With N=0, an allocation is temporary only if it
                is immediately deallocated before any other allocation occurs.
                """
            ),
            action="store",
            dest="temporary_allocation_threshold",
            type=int,
            default=-1,
        )
        alloc_type_group.add_argument(
            "--temporary-allocations",
            help="Equivalent to --temporary-allocation-threshold=1",
            action="store_const",
            dest="temporary_allocation_threshold",
            const=1,
        )

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        result_path = Path(args.results)
        if not result_path.exists() or not result_path.is_file():
            raise MemrayCommandError(f"No such file: {args.results}", exit_code=1)

        try:
            reader = FileReader(os.fspath(args.results), report_progress=True)
            if reader.metadata.has_native_traces:
                warn_if_not_enough_symbols()

            if not args.temporary_allocation_threshold >= 0:
                warn_if_file_is_not_aggregated_and_is_too_big(reader, result_path)

            if args.temporary_allocation_threshold >= 0:
                snapshot = iter(
                    reader.get_temporary_allocation_records(
                        threshold=args.temporary_allocation_threshold,
                        merge_threads=False,
                    )
                )
            else:
                snapshot = iter(
                    reader.get_high_watermark_allocation_records(merge_threads=False)
                )
            reporter = TreeReporter.from_snapshot(
                snapshot,
                biggest_allocs=args.biggest_allocs,
                native_traces=reader.metadata.has_native_traces,
            )
        except OSError as e:
            raise MemrayCommandError(
                f"Failed to parse allocation records in {result_path}\nReason: {e}",
                exit_code=1,
            )
        reporter.render()



from typing import Callable

from ._test_utils import MemoryAllocator as _MemoryAllocator
from ._test_utils import MmapAllocator
from ._test_utils import PrimeCaches
from ._test_utils import PymallocDomain
from ._test_utils import PymallocMemoryAllocator
from ._test_utils import _cython_allocate_in_two_places
from ._test_utils import _cython_nested_allocation
from ._test_utils import allocate_cpp_vector
from ._test_utils import allocate_without_gil_held
from ._test_utils import exit
from ._test_utils import fill_cpp_vector
from ._test_utils import function_caller
from ._test_utils import set_thread_name


class MemoryAllocator:
    def __init__(self) -> None:
        self.allocator = _MemoryAllocator()

    def free(self) -> None:
        return self.allocator.free()

    def malloc(self, size: int) -> bool:
        return self.allocator.malloc(size)

    def calloc(self, size: int) -> bool:
        return self.allocator.calloc(size)

    def realloc(self, size: int) -> bool:
        return self.allocator.realloc(size)

    def posix_memalign(self, size: int) -> bool:
        return self.allocator.posix_memalign(size)

    def aligned_alloc(self, size: int) -> bool:
        return self.allocator.aligned_alloc(size)

    def memalign(self, size: int) -> bool:
        return self.allocator.memalign(size)

    def valloc(self, size: int) -> bool:
        return self.allocator.valloc(size)

    def pvalloc(self, size: int) -> bool:
        return self.allocator.pvalloc(size)

    def run_in_pthread(self, callback: Callable[[], None]) -> None:
        return self.allocator.run_in_pthread(callback)


__all__ = [
    "allocate_cpp_vector",
    "MemoryAllocator",
    "MmapAllocator",
    "PymallocDomain",
    "PymallocMemoryAllocator",
    "_cython_allocate_in_two_places",
    "_cython_nested_allocation",
    "allocate_without_gil_held",
    "function_caller",
    "set_thread_name",
    "fill_cpp_vector",
    "exit",
    "PrimeCaches",
]



import sys
from typing import TextIO

from memray import Metadata

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


class BaseReporter(Protocol):
    def render(
        self,
        outfile: TextIO,
        metadata: Metadata,
        show_memory_leaks: bool,
        merge_threads: bool,
        inverted: bool,
        no_web: bool = False,
    ) -> None:
        ...



import datetime
import json
import math
from collections import Counter
from dataclasses import asdict
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import rich

from memray._memray import size_fmt
from memray._stats import Stats

PythonStackElement = Tuple[str, str, int]


def get_histogram_databins(data: Dict[int, int], bins: int) -> List[Tuple[int, int]]:
    if bins <= 0:
        raise ValueError(f"Invalid input bins={bins}, should be greater than 0")

    low = math.log(min(filter(None, data)))
    high = math.log(max(data))
    if low == high:
        low = low / 2
    step = (high - low) / bins

    # Determine the upper bound in bytes for each bin
    steps = [int(math.exp(low + step * (i + 1))) for i in range(bins)]
    dist: Dict[int, int] = Counter()
    for size, count in data.items():
        bucket = min(int((math.log(size) - low) // step), bins - 1) if size else 0
        dist[bucket] += count
    return [(steps[b], dist[b]) for b in range(bins)]


def describe_histogram_databins(
    databins: List[Tuple[int, int]]
) -> List[Dict[str, int]]:
    ret: List[Dict[str, int]] = []
    start = 0
    for i, (end, count) in enumerate(databins):
        # The max size for the last bucket is inclusive, not exclusive
        adjustment = 1 if i != len(databins) - 1 else 0
        ret.append({"min_bytes": start, "max_bytes": end - adjustment, "count": count})
        start = end
    return ret


def draw_histogram(
    data: Dict[int, int], bins: int, *, hist_scale_factor: int = 25
) -> str:
    """
    @param data: list of allocation sizes
    @param bins: number of bins in the histogram
    @param hist_scale_factor: length of the largest bar in the histogram (# of chars)
    """
    if len(data) == 0:
        return "<no data for histogram>"
    if bins <= 0:
        raise ValueError(f"Invalid input bins={bins}, should be greater than 0")
    if hist_scale_factor <= 0:
        raise ValueError(
            f"Invalid input hist_scale_factor={hist_scale_factor},"
            " should be greater than 0"
        )

    data_bins = get_histogram_databins(data, bins=bins)
    max_data_bin = max([t[1] for t in data_bins])
    scaled_data_bins = [
        math.ceil((v / max_data_bin) * hist_scale_factor) for _, v in data_bins
    ]

    size_max_width = 9  # ###.###XB
    freq_max_width = max([len(str(f[1])) for f in data_bins])
    hist_width_total = (
        len("<=")
        + size_max_width
        + len(" :")
        + freq_max_width
        + len(" ")
        + hist_scale_factor
    )

    result = []
    result.append(f"min: {size_fmt(min(data))}")
    result.append("\n\t")
    result.append("-" * hist_width_total)
    result.append("\n\t")
    for i in range(len(scaled_data_bins)):
        rel_op = "<=" if i == len(scaled_data_bins) - 1 else "< "
        result.append(f"{rel_op}{size_fmt(data_bins[i][0]):<{size_max_width}}:")
        result.append(f" {data_bins[i][1]:>{freq_max_width}} ")
        result.append("▇" * scaled_data_bins[i])
        result.append("\n\t")
    result.append("-" * hist_width_total)
    result.append(f"\n\tmax: {size_fmt(max(data))}")

    return "".join(result)


class StatsReporter:
    def __init__(self, stats: Stats, num_largest: int):
        self._stats = stats
        if num_largest < 1:
            raise ValueError(f"Invalid input num_largest={num_largest}, should be >=1")
        self.num_largest = num_largest

    def render(self, json_output_file: Optional[Path] = None) -> None:
        histogram_params = {
            "num_bins": 10,
            "histogram_scale_factor": 25,
        }
        if json_output_file:
            self._render_to_json(histogram_params, json_output_file)
        else:
            self._render_to_terminal(histogram_params)

    def _render_to_terminal(self, histogram_params: Dict[str, int]) -> None:
        rich.print("📏 [bold]Total allocations:[/]")
        print(f"\t{self._stats.total_num_allocations}")

        print()
        rich.print("📦 [bold]Total memory allocated:[/]")
        print(f"\t{size_fmt(self._stats.total_memory_allocated)}")

        print()
        rich.print("📈 [bold]Peak memory usage:[/]")
        print(f"\t{size_fmt(self._stats.metadata.peak_memory)}")

        print()

        rich.print("📊 [bold]Histogram of allocation size:[/]")
        histogram = draw_histogram(
            self._stats.allocation_count_by_size,
            histogram_params["num_bins"],
            hist_scale_factor=histogram_params["histogram_scale_factor"],
        )
        print(f"\t{histogram}")

        print()
        rich.print("📂 [bold]Allocator type distribution:[/]")
        for allocator_name, count in self._get_allocator_type_distribution():
            print(f"\t {allocator_name}: {count}")

        print()
        rich.print(
            f"🥇 [bold]Top {self.num_largest} largest allocating locations (by size):[/]"
        )
        for location, size in self._get_top_allocations_by_size():
            print(f"\t- {self._format_location(location)} -> {size_fmt(size)}")

        print()
        rich.print(
            f"🥇 [bold]Top {self.num_largest} largest allocating "
            "locations (by number of allocations):[/]"
        )
        for location, count in self._get_top_allocations_by_count():
            print(f"\t- {self._format_location(location)} -> {count}")

    def _render_to_json(self, histogram_params: Dict[str, int], out_path: Path) -> None:
        alloc_size_hist = describe_histogram_databins(
            get_histogram_databins(
                self._stats.allocation_count_by_size, bins=histogram_params["num_bins"]
            )
        )

        metadata = asdict(self._stats.metadata)
        for name, val in metadata.items():
            if isinstance(val, datetime.datetime):
                metadata[name] = str(val)

        data: Dict[str, Any] = {
            "total_num_allocations": self._stats.total_num_allocations,
            "total_bytes_allocated": self._stats.total_memory_allocated,
            "allocation_size_histogram": alloc_size_hist,
            "allocator_type_distribution": dict(
                self._get_allocator_type_distribution()
            ),
            "top_allocations_by_size": [
                {"location": self._format_location(location), "size": size}
                for location, size in self._get_top_allocations_by_size()
            ],
            "top_allocations_by_count": [
                {"location": self._format_location(location), "count": count}
                for location, count in self._get_top_allocations_by_count()
            ],
            "metadata": metadata,
        }

        with open(out_path, "w") as f:
            json.dump(data, f, indent=2)

    @staticmethod
    def _format_location(loc: Tuple[str, str, int]) -> str:
        function, file, line = loc
        if function == "<unknown>":
            return "<stack trace unavailable>"
        return f"{function}:{file}:{line}"

    def _get_top_allocations_by_size(self) -> Iterator[Tuple[PythonStackElement, int]]:
        for location, size in self._stats.top_locations_by_size:
            yield (location, size)

    def _get_top_allocations_by_count(self) -> Iterator[Tuple[PythonStackElement, int]]:
        for location, count in self._stats.top_locations_by_count:
            yield (location, count)

    def _get_allocator_type_distribution(self) -> Iterator[Tuple[str, int]]:
        for allocator_name, count in sorted(
            self._stats.allocation_count_by_allocator.items(),
            key=lambda item: item[1],
            reverse=True,
        ):
            yield (allocator_name, count)



from dataclasses import dataclass

from ._metadata import Metadata


@dataclass
class Stats:
    metadata: Metadata
    total_num_allocations: int
    total_memory_allocated: int
    peak_memory_allocated: int
    allocation_count_by_size: dict
    allocation_count_by_allocator: dict
    top_locations_by_size: list
    top_locations_by_count: list



from typing import Union

from memray._memray import AllocationRecord
from memray._memray import TemporalAllocationRecord


def format_thread_name(
    record: Union[AllocationRecord, TemporalAllocationRecord]
) -> str:
    if record.tid == -1:
        return "merged thread"
    name = record.thread_name
    thread_id = hex(record.tid)
    return f"{thread_id} ({name})" if name else f"{thread_id}"



from ._ipython import load_ipython_extension
from ._memray import AllocationRecord
from ._memray import AllocatorType
from ._memray import Destination
from ._memray import FileDestination
from ._memray import FileFormat
from ._memray import FileReader
from ._memray import MemorySnapshot
from ._memray import SocketDestination
from ._memray import SocketReader
from ._memray import Tracker
from ._memray import dump_all_records
from ._memray import set_log_level
from ._memray import start_thread_trace
from ._metadata import Metadata
from ._version import __version__

__all__ = [
    "AllocationRecord",
    "AllocatorType",
    "FileFormat",
    "MemorySnapshot",
    "dump_all_records",
    "start_thread_trace",
    "Tracker",
    "FileReader",
    "SocketReader",
    "Destination",
    "FileDestination",
    "SocketDestination",
    "Metadata",
    "__version__",
    "set_log_level",
    "load_ipython_extension",
]



import dataclasses
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Union

from textual import binding
from textual.app import App
from textual.binding import Binding
from textual.dom import DOMNode
from textual.widgets import Footer

# In Textual 0.61, `App.namespace_bindings` was removed in favor of
# `Screen.active_bindings`. The two have a slightly different interface:
# a 2 item `tuple` was updated to a 3 item `namedtuple`.
# The `Bindings` type alias shows the two possible structures.
# The `update_key_description` implementation works for both,
# since we still support Textual versions below 0.61.

Bindings = Union[Dict[str, "binding.ActiveBinding"], Dict[str, Tuple[DOMNode, Binding]]]


def update_key_description(bindings: Bindings, key: str, description: str) -> None:
    val = bindings[key]
    binding = dataclasses.replace(val[1], description=description)
    if type(val) is tuple:
        bindings[key] = val[:1] + (binding,) + val[2:]  # type: ignore
    else:
        bindings[key] = val._replace(binding=binding)  # type: ignore


def redraw_footer(app: App[Any]) -> None:
    footer = app.screen.query_one(Footer)
    if hasattr(footer, "recompose"):
        # Added in Textual v0.53
        footer.refresh(recompose=True)
    else:  # pragma: no cover
        # Hack: trick the Footer into redrawing itself
        footer.highlight_key = "q"  # type: ignore[attr-defined]
        footer.highlight_key = None  # type: ignore[attr-defined]



import argparse

from ..reporters.flamegraph import FlameGraphReporter
from .common import HighWatermarkCommand


class FlamegraphCommand(HighWatermarkCommand):
    """Generate an HTML flame graph for peak memory usage"""

    def __init__(self) -> None:
        super().__init__(
            reporter_factory=FlameGraphReporter.from_snapshot,
            temporal_reporter_factory=FlameGraphReporter.from_temporal_snapshot,
            reporter_name="flamegraph",
        )

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        super().prepare_parser(parser)
        parser.add_argument(
            "--split-threads",
            help="Do not merge allocations across threads",
            action="store_true",
            default=False,
        )

        parser.add_argument(
            "--inverted",
            help="Invert flame graph",
            action="store_true",
            default=False,
        )

        parser.add_argument(
            "--max-memory-records",
            help="Maximum number of memory records to display",
            type=int,
            default=None,
        )

        parser.add_argument(
            "--no-web",
            help="Use local assets instead of fetching from CDN",
            action="store_true",
            default=False,
        )



import argparse
import importlib.util
import shutil
import sys

from rich import print as pprint

from memray._errors import MemrayCommandError

from ..reporters.transform import TransformReporter
from .common import HighWatermarkCommand


class TransformCommand(HighWatermarkCommand):
    """Generate reports files in different formats"""

    def __init__(self) -> None:
        super().__init__(
            reporter_factory=lambda *args, **kwargs: TransformReporter(
                *args, **kwargs, format=self.reporter_name
            ),
            reporter_name="transform",
        )

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        formats = ", ".join(TransformReporter.SUFFIX_MAP)
        parser.add_argument(
            "format",
            help=f"Format to use for the report. Available formats: {formats}",
        )
        super().prepare_parser(parser)

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        the_format = args.format.lower()
        suffix = TransformReporter.SUFFIX_MAP.get(the_format)
        if not suffix:
            raise MemrayCommandError(
                f"Format not supported: {args.format}", exit_code=1
            )

        self.suffix = suffix
        self.reporter_name = the_format
        super().run(args, parser)

        post_run_callable = getattr(self, f"post_run_{the_format}", None)
        if post_run_callable:
            post_run_callable()

    def post_run_gprof2dot(self) -> None:
        assert self.output_file is not None
        command = ""
        if shutil.which("gprof2dot") is not None:
            command = "gprof2dot"
        elif importlib.util.find_spec("gprof2dot") is not None:
            command = f"{sys.executable} -m gprof2dot"
        else:
            pprint(
                ":exclamation: [red]gprof2dot doesn't seem to be installed. "
                "Please install it to be able to process the transform.[/red]"
                "\n:point_right: Check out https://github.com/jrfonseca/gprof2dot for "
                "installation instructions"
            )

        if not shutil.which("dot"):
            pprint(
                ":exclamation: [red]Graphviz doesn't seem to be installed. "
                "Please install it to be able to process the transform.[/red]"
                "\n:point_right: Check out https://graphviz.org/download/ for "
                "installation instructions"
            )

        print()
        print("To generate a graph from the transform file, run for example:")
        print(f"{command} -f json {self.output_file} | dot -Tpng -o output.png")



from memray._destination import Destination as Destination
from memray._destination import FileDestination as FileDestination
from memray._destination import SocketDestination as SocketDestination
from memray._metadata import Metadata as Metadata

from ._memray import AllocationRecord as AllocationRecord
from ._memray import AllocatorType as AllocatorType
from ._memray import FileFormat as FileFormat
from ._memray import FileReader as FileReader
from ._memray import MemorySnapshot as MemorySnapshot
from ._memray import SocketReader as SocketReader
from ._memray import Tracker as Tracker
from ._memray import dump_all_records as dump_all_records



import asyncio
import functools
import linecache
import sys
from dataclasses import dataclass
from dataclasses import field
from typing import IO
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import Optional
from typing import Tuple

from rich.style import Style
from rich.text import Text
from textual import binding
from textual import work
from textual.app import App
from textual.app import ComposeResult
from textual.binding import Binding
from textual.color import Color
from textual.color import Gradient
from textual.containers import Grid
from textual.containers import Horizontal
from textual.containers import Vertical
from textual.dom import DOMNode
from textual.reactive import reactive
from textual.screen import Screen
from textual.widget import Widget
from textual.widgets import Footer
from textual.widgets import Label
from textual.widgets import TextArea
from textual.widgets import Tree
from textual.widgets.tree import TreeNode

from memray import AllocationRecord
from memray._memray import size_fmt
from memray.reporters._textual_hacks import Bindings
from memray.reporters._textual_hacks import redraw_footer
from memray.reporters._textual_hacks import update_key_description
from memray.reporters.common import format_thread_name
from memray.reporters.frame_tools import is_cpython_internal
from memray.reporters.frame_tools import is_frame_from_import_system
from memray.reporters.frame_tools import is_frame_interesting
from memray.reporters.tui import _filename_to_module_name

MAX_STACKS = int(sys.getrecursionlimit() // 2.5)

StackElement = Tuple[str, str, int]

ROOT_NODE = ("<ROOT>", "", 0)


@dataclass
class Frame:
    """A frame in the tree"""

    location: Optional[StackElement]
    value: int
    children: Dict[StackElement, "Frame"] = field(default_factory=dict)
    n_allocations: int = 0
    thread_id: str = ""
    interesting: bool = True
    import_system: bool = False


@dataclass
class ElidedLocations:
    """Information about allocations locations below the configured threshold."""

    cutoff: int = 0
    n_locations: int = 0
    n_allocations: int = 0
    n_bytes: int = 0


class FrameDetailScreen(Widget):
    """A screen that displays information about a frame"""

    frame = reactive(Frame(location=ROOT_NODE, value=0))

    def __init__(
        self, *args: Any, elided_locations: ElidedLocations, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.__elided_locations = elided_locations
        self.__is_mounted = False

    def on_mount(self) -> None:
        self.__is_mounted = True

    @work(exclusive=True)
    async def update_text_area(self) -> None:
        await asyncio.sleep(0.1)

        if not self.__is_mounted or self.frame is None:
            return

        text = self.query_one("#textarea", TextArea)

        if self.frame.location is None or self.frame.location == ROOT_NODE:
            text.clear()
            return

        _, file, line = self.frame.location
        delta = text.size.height // 2
        lines = linecache.getlines(file)[max(line - delta, 0) : line + delta]

        text.text = "\n".join(tuple(line.rstrip() for line in lines))
        text.select_line(line - 1 if delta >= line else delta - 1)
        text.show_line_numbers = False

    def _get_content_by_label_id(self) -> Dict[str, str]:
        common = {
            "allocs": f"\N{floppy disk} Allocations: {self.frame.n_allocations}",
            "size": f"\N{package} Size: {size_fmt(self.frame.value)}",
        }

        if self.frame.location is None:
            cutoff = self.__elided_locations.cutoff
            return {
                **common,
                "function": "",
                "location": (
                    f"Only the top {cutoff} allocation locations are shown in the tree."
                    " Allocation locations which individually contributed too little"
                    " to meet the threshold are summarized here.\n\n"
                    "You can adjust this threshold to include more allocation locations"
                    " by rerunning this reporter with a larger --biggest-allocs value."
                ),
                "thread": "",
            }

        function, file, lineno = self.frame.location
        if self.frame.location is ROOT_NODE:
            return {
                **common,
                "function": "",
                "location": "",
                "thread": "",
            }
        return {
            **common,
            "function": f"\N{compass} Function: {function}",
            "location": (
                "\N{compass} Location: "
                + (
                    f"{_filename_to_module_name(file)}:{lineno}"
                    if lineno != 0
                    else file
                )
            ),
            "thread": f"\N{spool of thread} Thread: {self.frame.thread_id}",
        }

    def watch_frame(self) -> None:
        if not self.__is_mounted or self.frame is None:
            return

        self.update_text_area()

        content_by_label_id = self._get_content_by_label_id()
        for label_id, content in content_by_label_id.items():
            label = self.query_one(f"#{label_id}", Label)
            label.update(content)
            label.set_class(not content, "hidden")
            label.styles.display = "block" if content else "none"

    def compose(self) -> ComposeResult:
        if self.frame is None:
            return

        delta = 3

        if self.frame.location is None or self.frame.location == ROOT_NODE:
            lines = []
            selected_line = 0
        else:
            _, file, line = self.frame.location
            lines = linecache.getlines(file)[max(line - delta, 0) : line + delta]
            selected_line = line - 1 if delta >= line else delta - 1

        text = TextArea(
            "\n".join(lines), language="python", theme="dracula", id="textarea"
        )
        text.select_line(selected_line)
        text.show_line_numbers = False
        text.can_focus = False
        text.cursor_blink = False
        text.soft_wrap = False

        labels: list[Label] = []
        content_by_label_id = self._get_content_by_label_id()
        for label_id in ("function", "location", "allocs", "size", "thread"):
            content = content_by_label_id[label_id]
            label = Label(content, id=label_id)
            label.styles.display = "block" if content else "none"
            labels.append(label)

        node_metadata = Vertical(*labels)
        yield Grid(
            text,
            node_metadata,
            id="frame-detail-grid",
        )


class FrameTree(Tree[Frame]):
    def on_tree_node_selected(self, node: Tree.NodeSelected[Frame]) -> None:
        if node.node.data is not None:
            self.app.screen.query_one(FrameDetailScreen).frame = node.node.data

    def on_tree_node_highlighted(self, node: Tree.NodeHighlighted[Frame]) -> None:
        if node.node.data is not None:
            self.app.screen.query_one(FrameDetailScreen).frame = node.node.data


def node_is_interesting(node: Frame) -> bool:
    return node.interesting


def node_is_not_import_system(node: Frame) -> bool:
    return not node.import_system


class TreeScreen(Screen[None]):
    BINDINGS = [
        Binding("ctrl+z", "app.suspend_process"),
        Binding(key="q", action="app.quit", description="Quit the app"),
        Binding(
            key="i", action="toggle_import_system", description="Hide import system"
        ),
        Binding(
            key="u", action="toggle_uninteresting", description="Hide uninteresting"
        ),
        Binding(
            key="e", action="expand_linear_group", description="Expand linear group"
        ),
    ]

    CSS_PATH = "tree.css"

    def __init__(
        self,
        data: Frame,
        elided_locations: ElidedLocations,
    ):
        super().__init__()
        self.data = data
        self.elided_locations = elided_locations
        self.import_system_filter: Optional[Callable[[Frame], bool]] = None
        self.uninteresting_filter: Optional[
            Callable[[Frame], bool]
        ] = node_is_interesting

    def expand_first_child(self, node: TreeNode[Frame]) -> None:
        while node.children:
            node = node.children[0]
            node.toggle()

    def compose(self) -> ComposeResult:
        tree = FrameTree(self.frame_text(self.data, allow_expand=True), self.data)
        self.repopulate_tree(tree)
        yield Horizontal(
            Vertical(tree),
            Vertical(
                FrameDetailScreen(elided_locations=self.elided_locations),
                id="detailcol",
            ),
        )
        yield Footer()

    def repopulate_tree(self, tree: FrameTree) -> None:
        tree.clear()
        self.add_children(tree.root, self.data.children.values())
        self.add_elided_locations_node(tree.root)
        tree.root.expand()
        # From Textual 0.73 on, Tree.select_node toggles the node's expanded
        # state. The new Tree.move_cursor method selects without expanding.
        getattr(tree, "move_cursor", tree.select_node)(tree.root)
        self.expand_first_child(tree.root)

    def action_expand_linear_group(self) -> None:
        tree = self.query_one(FrameTree)
        current_node = tree.cursor_node
        while current_node:
            current_node.toggle()
            if len(current_node.children) != 1:
                break
            current_node = current_node.children[0]

    def frame_text(self, node: Frame, *, allow_expand: bool) -> Text:
        if node.value == 0:
            return Text("<No allocations>")

        value = node.value
        root_data = self.data
        size_str = f"{size_fmt(value)} ({100 * value / root_data.value:.2f} %)"
        size_color = _info_color(node, root_data)

        ret = Text.from_markup(
            ":open_file_folder:" if allow_expand else ":page_facing_up:"
        )
        ret.append_text(Text(f" {size_str} ", style=Style(color=size_color.rich_color)))

        if node.location is not None:
            function, file, lineno = node.location
            code_position = (
                f"{_filename_to_module_name(file)}:{lineno}" if lineno != 0 else file
            )
            ret.append_text(Text.from_markup(f"[bold]{function}[/]"))
            if code_position:
                ret.append_text(Text.from_markup(f"  [dim cyan]{code_position}[/]"))
        else:
            ret.append_text(Text("hidden"))
        return ret

    def add_children(self, tree: TreeNode[Frame], children: Iterable[Frame]) -> None:
        # Add children to the tree from largest to smallest
        children = sorted(children, key=lambda child: child.value, reverse=True)

        if self.import_system_filter is not None:
            children = tuple(filter(self.import_system_filter, children))

        for child in children:
            if self.uninteresting_filter is None or self.uninteresting_filter(child):
                if not tree.allow_expand:
                    assert tree.data is not None
                    tree.label = self.frame_text(tree.data, allow_expand=True)
                    tree.allow_expand = True
                new_tree = tree.add(
                    self.frame_text(child, allow_expand=False),
                    data=child,
                    allow_expand=False,
                )
            else:
                new_tree = tree

            self.add_children(new_tree, child.children.values())

    def add_elided_locations_node(self, tree: TreeNode[Frame]) -> None:
        if not self.elided_locations.n_locations:
            return

        count = self.elided_locations.n_locations
        value = self.elided_locations.n_bytes
        number = self.elided_locations.n_allocations

        root_data = self.data
        percentage = 100 * value / root_data.value
        size_str = f"{size_fmt(value)} ({percentage:.2f} %)"
        size_color = _percentage_to_color(int(percentage))
        ret = Text.from_markup("\N{black question mark ornament}")
        ret.append_text(Text(f" {size_str} ", style=Style(color=size_color.rich_color)))
        ret.append_text(
            Text.from_markup(
                f"{number} allocations from {count} locations"
                f" below the configured threshold"
            )
        )

        tree.add_leaf(ret, data=Frame(location=None, value=value, n_allocations=number))

    def action_toggle_import_system(self) -> None:
        if self.import_system_filter is None:
            self.import_system_filter = node_is_not_import_system
        else:
            self.import_system_filter = None

        redraw_footer(self.app)
        self.repopulate_tree(self.query_one(FrameTree))

    def action_toggle_uninteresting(self) -> None:
        if self.uninteresting_filter is None:
            self.uninteresting_filter = node_is_interesting
        else:
            self.uninteresting_filter = None

        redraw_footer(self.app)
        self.repopulate_tree(self.query_one(FrameTree))

    def rewrite_bindings(self, bindings: Bindings) -> None:
        if self.import_system_filter is not None:
            update_key_description(bindings, "i", "Show import system")
        if self.uninteresting_filter is not None:
            update_key_description(bindings, "u", "Show uninteresting")

    @property
    def active_bindings(self) -> Dict[str, "binding.ActiveBinding"]:
        bindings = super().active_bindings.copy()
        self.rewrite_bindings(bindings)
        return bindings


class TreeApp(App[None]):
    def __init__(
        self,
        data: Frame,
        elided_locations: ElidedLocations,
    ):
        super().__init__()
        self.tree_screen = TreeScreen(data, elided_locations)

    def on_mount(self) -> None:
        self.push_screen(self.tree_screen)

    if hasattr(App, "namespace_bindings"):
        # Removed in Textual 0.61
        @property
        def namespace_bindings(self) -> Dict[str, Tuple[DOMNode, Binding]]:
            bindings = super().namespace_bindings.copy()  # type: ignore[misc]
            self.tree_screen.rewrite_bindings(bindings)
            return bindings  # type: ignore[no-any-return]


@functools.lru_cache(maxsize=None)
def _percentage_to_color(percentage: int) -> Color:
    gradient = Gradient(
        (0, Color(97, 193, 44)),
        (0.4, Color(236, 152, 16)),
        (0.6, Color.parse("darkorange")),
        (1, Color.parse("indianred")),
    )
    return gradient.get_color(percentage / 100)


def _info_color(node: Frame, root_node: Frame) -> Color:
    proportion_of_total = node.value / root_node.value
    return _percentage_to_color(int(proportion_of_total * 100))


class TreeReporter:
    def __init__(self, data: Frame, elided_locations: ElidedLocations) -> None:
        super().__init__()
        self.data = data
        self.elided_locations = elided_locations

    @classmethod
    def from_snapshot(
        cls,
        allocations: Iterator[AllocationRecord],
        *,
        biggest_allocs: int = 200,
        native_traces: bool,
    ) -> "TreeReporter":
        data = Frame(location=ROOT_NODE, value=0, import_system=False, interesting=True)
        sorted_records = sorted(allocations, key=lambda alloc: alloc.size, reverse=True)
        for record in sorted_records[:biggest_allocs]:
            size = record.size
            data.value += size
            data.n_allocations += record.n_allocations

            current_frame = data
            stack = (
                tuple(record.hybrid_stack_trace())
                if native_traces
                else record.stack_trace()
            )
            for index, stack_frame in enumerate(reversed(stack)):
                if is_cpython_internal(stack_frame):
                    continue
                is_import_system = is_frame_from_import_system(stack_frame)
                is_interesting = not is_import_system and is_frame_interesting(
                    stack_frame
                )
                if stack_frame not in current_frame.children:
                    node = Frame(
                        value=0,
                        location=stack_frame,
                        import_system=is_import_system,
                        interesting=is_interesting,
                    )
                    current_frame.children[stack_frame] = node

                current_frame = current_frame.children[stack_frame]
                current_frame.value += size
                current_frame.n_allocations += record.n_allocations
                current_frame.thread_id = format_thread_name(record)

                if index > MAX_STACKS:
                    break

        elided_locations = ElidedLocations()
        elided_locations.cutoff = biggest_allocs

        for record in sorted_records[biggest_allocs:]:
            data.value += record.size
            data.n_allocations += record.n_allocations
            elided_locations.n_locations += 1
            elided_locations.n_bytes += record.size
            elided_locations.n_allocations += record.n_allocations

        return cls(data, elided_locations)

    def get_app(self) -> TreeApp:
        return TreeApp(self.data, self.elided_locations)

    def render(
        self,
        *,
        file: Optional[IO[str]] = None,
    ) -> None:
        self.get_app().run()



from __future__ import annotations

import typing
from dataclasses import dataclass
from datetime import datetime

if typing.TYPE_CHECKING:
    from ._memray import FileFormat


@dataclass
class Metadata:
    start_time: datetime
    end_time: datetime
    total_allocations: int
    total_frames: int
    peak_memory: int
    command_line: str
    pid: int
    main_thread_id: int
    python_allocator: str
    has_native_traces: bool
    trace_python_allocators: bool
    file_format: FileFormat



import csv
import json
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import TextIO
from typing import Tuple

from memray import AllocationRecord
from memray import AllocatorType
from memray import MemorySnapshot
from memray import Metadata
from memray.reporters.common import format_thread_name

Location = Tuple[str, str]


class TransformReporter:
    SUFFIX_MAP = {
        "gprof2dot": ".json",
        "csv": ".csv",
    }

    def __init__(
        self,
        allocations: Iterable[AllocationRecord],
        *,
        format: str,
        native_traces: bool,
        memory_records: Iterable[MemorySnapshot],
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.allocations = allocations
        self.format = format
        self.native_traces = native_traces
        self.memory_records = memory_records

    def render_as_gprof2dot(
        self,
        outfile: TextIO,
        **kwargs: Any,
    ) -> None:
        location_to_index: Dict[Location, int] = {}
        all_locations: List[Dict[str, str]] = []
        events = []
        for record in self.allocations:
            stack_trace = (
                tuple(record.hybrid_stack_trace())
                if self.native_traces
                else record.stack_trace()
            )
            call_chain = []
            for func, mod, _ in stack_trace:
                location = (func, mod)
                index = location_to_index.get(location)
                if index is None:
                    index = len(all_locations)
                    all_locations.append({"name": func, "module": mod})
                    location_to_index[location] = index
                call_chain.append(index)

            if not call_chain:
                continue
            events.append({"callchain": call_chain, "cost": [record.size]})

        result = {
            "version": 0,
            "costs": [{"description": "Memory", "unit": "bytes"}],
            "events": events,
            "functions": all_locations,
        }
        json.dump(result, outfile)

    def render(
        self,
        outfile: TextIO,
        metadata: Metadata,
        show_memory_leaks: bool,
        merge_threads: bool,
        inverted: bool,
        no_web: bool = False,
    ) -> None:
        if not merge_threads:
            raise NotImplementedError("TransformReporter only supports merged threads.")
        if inverted:
            raise NotImplementedError(
                "TransformReporter does not support inverted argument."
            )
        renderer = getattr(self, f"render_as_{self.format}")
        renderer(outfile, metadata=metadata, show_memory_leaks=show_memory_leaks)

    def render_as_csv(
        self,
        outfile: TextIO,
        **kwargs: Any,
    ) -> None:
        writer = csv.writer(outfile)
        writer.writerow(
            [
                "allocator",
                "num_allocations",
                "size",
                "tid",
                "thread_name",
                "stack_trace",
            ]
        )
        for record in self.allocations:
            stack_trace = (
                tuple(record.hybrid_stack_trace())
                if self.native_traces
                else record.stack_trace()
            )
            writer.writerow(
                [
                    AllocatorType(record.allocator).name,
                    record.n_allocations,
                    record.size,
                    record.tid,
                    format_thread_name(record),
                    "|".join(f"{func};{mod};{line}" for func, mod, line in stack_trace),
                ]
            )



import argparse
import os
import pathlib
import sys
from pathlib import Path
from textwrap import dedent
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

from rich import print as pprint

from memray import AllocationRecord
from memray import FileReader
from memray import MemorySnapshot
from memray._errors import MemrayCommandError
from memray._memray import FileFormat
from memray._memray import SymbolicSupport
from memray._memray import TemporalAllocationRecord
from memray._memray import get_symbolic_support
from memray.reporters import BaseReporter


class ReporterFactory(Protocol):
    def __call__(
        self,
        allocations: Iterable[AllocationRecord],
        *,
        memory_records: Iterable[MemorySnapshot],
        native_traces: bool,
        inverted: bool,
    ) -> BaseReporter:
        ...


class TemporalReporterFactory(Protocol):
    def __call__(
        self,
        allocations: Iterable[TemporalAllocationRecord],
        *,
        memory_records: Iterable[MemorySnapshot],
        native_traces: bool,
        high_water_mark_by_snapshot: Optional[List[int]],
        inverted: bool,
    ) -> BaseReporter:
        ...


def warn_if_not_enough_symbols() -> None:
    support = get_symbolic_support()
    if support == SymbolicSupport.NONE:
        pprint(
            ":warning: [bold yellow] No symbol information was found for the "
            "Python interpreter [/] :warning:\n\n"
            "Without symbolic information reports showing native traces [b]may not "
            "accurately reflect stack traces[/]. Please use an interpreter built with "
            "debug symbols for best results. Check "
            "https://bloomberg.github.io/memray/native_mode.html for more information "
            "regarding how memray resolves symbols.\n"
        )
    elif support == SymbolicSupport.FUNCTION_NAME_ONLY:
        pprint(
            ":warning: [bold yellow] No debug information was found for the "
            "Python interpreter [/] :warning:\n\n"
            "Without debug information reports showing native traces [b]may not "
            "include file names and line numbers[/]. Please use an interpreter built with "
            "debug symbols for best results. Check "
            "https://bloomberg.github.io/memray/native_mode.html for more information "
            "regarding how memray resolves symbols.\n"
        )
    else:
        return


def warn_if_file_is_not_aggregated_and_is_too_big(
    reader: FileReader, result_path: Path
) -> None:
    FILE_SIZE_LIMIT = 10 * 1000 * 1000
    if (
        reader.metadata.file_format == FileFormat.ALL_ALLOCATIONS
        and result_path.stat().st_size > FILE_SIZE_LIMIT
    ):
        pprint(
            ":warning: [bold yellow] This capture file is large and may take a long"
            " time to process [/] :warning:\n\n"
            "Next time, consider using the `--aggregate` option to `memray run` to"
            " reduce the size of the file.\n"
            "Check https://bloomberg.github.io/memray/run.html#aggregated-capture-files"
            " for more information.\n"
        )


class HighWatermarkCommand:
    def __init__(
        self,
        reporter_factory: ReporterFactory,
        reporter_name: str,
        suffix: str = ".html",
        temporal_reporter_factory: Optional[TemporalReporterFactory] = None,
    ) -> None:
        self.reporter_factory = reporter_factory
        self.reporter_name = reporter_name
        self.suffix = suffix
        self.temporal_reporter_factory = temporal_reporter_factory
        self.output_file: Optional[Path] = None

    def determine_output_filename(self, results_file: pathlib.Path) -> pathlib.Path:
        output_name = results_file.with_suffix(self.suffix).name
        if output_name.startswith("memray-"):
            output_name = output_name[len("memray-") :]

        return results_file.parent / f"memray-{self.reporter_name}-{output_name}"

    def validate_filenames(
        self, output: Optional[str], results: str, overwrite: bool = False
    ) -> Tuple[Path, Path]:
        """Ensure that the filenames provided by the user are usable."""
        result_path = Path(results)
        if not result_path.exists() or not result_path.is_file():
            raise MemrayCommandError(f"No such file: {results}", exit_code=1)

        output_file = Path(
            output
            if output is not None
            else self.determine_output_filename(result_path)
        )
        if not overwrite and output_file.exists():
            raise MemrayCommandError(
                f"File already exists, will not overwrite: {output_file}",
                exit_code=1,
            )
        return result_path, output_file

    def write_report(
        self,
        result_path: Path,
        output_file: Path,
        show_memory_leaks: bool,
        temporary_allocation_threshold: int,
        merge_threads: Optional[bool] = None,
        inverted: Optional[bool] = None,
        temporal: bool = False,
        max_memory_records: Optional[int] = None,
        no_web: bool = False,
    ) -> None:
        try:
            kwargs = {}
            if max_memory_records is not None:
                kwargs["max_memory_records"] = max_memory_records
            reader = FileReader(os.fspath(result_path), report_progress=True, **kwargs)
            merge_threads = True if merge_threads is None else merge_threads
            inverted = False if inverted is None else inverted

            if reader.metadata.has_native_traces:
                warn_if_not_enough_symbols()

            if not temporal and not temporary_allocation_threshold >= 0:
                warn_if_file_is_not_aggregated_and_is_too_big(reader, result_path)

            if temporal:
                assert self.temporal_reporter_factory is not None
                if show_memory_leaks:
                    temporal_snapshot = reader.get_temporal_allocation_records(
                        merge_threads=merge_threads
                    )
                    reporter = self.temporal_reporter_factory(
                        temporal_snapshot,
                        memory_records=tuple(reader.get_memory_snapshots()),
                        native_traces=reader.metadata.has_native_traces,
                        high_water_mark_by_snapshot=None,
                        inverted=inverted,
                    )
                else:
                    recs, hwms = reader.get_temporal_high_water_mark_allocation_records(
                        merge_threads=merge_threads
                    )
                    reporter = self.temporal_reporter_factory(
                        recs,
                        memory_records=tuple(reader.get_memory_snapshots()),
                        native_traces=reader.metadata.has_native_traces,
                        high_water_mark_by_snapshot=hwms,
                        inverted=inverted,
                    )
            else:
                if show_memory_leaks:
                    snapshot = reader.get_leaked_allocation_records(
                        merge_threads=merge_threads
                    )
                elif temporary_allocation_threshold >= 0:
                    snapshot = reader.get_temporary_allocation_records(
                        threshold=temporary_allocation_threshold,
                        merge_threads=merge_threads,
                    )
                else:
                    snapshot = reader.get_high_watermark_allocation_records(
                        merge_threads=merge_threads
                    )
                reporter = self.reporter_factory(
                    snapshot,
                    memory_records=tuple(reader.get_memory_snapshots()),
                    native_traces=reader.metadata.has_native_traces,
                    inverted=inverted,
                )
        except OSError as e:
            raise MemrayCommandError(
                f"Failed to parse allocation records in {result_path}\nReason: {e}",
                exit_code=1,
            )

        with open(os.fspath(output_file.expanduser()), "w") as f:
            reporter.render(
                outfile=f,
                metadata=reader.metadata,
                show_memory_leaks=show_memory_leaks,
                merge_threads=merge_threads,
                inverted=inverted,
                no_web=no_web,
            )

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-o",
            "--output",
            help="Output file name",
            default=None,
        )
        parser.add_argument(
            "-f",
            "--force",
            help="If the output file already exists, overwrite it",
            action="store_true",
            default=False,
        )
        if self.temporal_reporter_factory:
            parser.add_argument(
                "--temporal",
                help=(
                    "Generate a dynamic flame graph that can analyze"
                    " allocations in a user-selected time range."
                ),
                action="store_true",
                default=False,
            )

        alloc_type_group = parser.add_mutually_exclusive_group()
        alloc_type_group.add_argument(
            "--leaks",
            help="Show memory leaks, instead of peak memory usage",
            action="store_true",
            dest="show_memory_leaks",
            default=False,
        )
        alloc_type_group.add_argument(
            "--temporary-allocation-threshold",
            metavar="N",
            help=dedent(
                """
                Report temporary allocations, as opposed to leaked allocations
                or high watermark allocations.  An allocation is considered
                temporary if at most N other allocations occur before it is
                deallocated.  With N=0, an allocation is temporary only if it
                is immediately deallocated before any other allocation occurs.
                """
            ),
            action="store",
            dest="temporary_allocation_threshold",
            type=int,
            default=-1,
        )
        alloc_type_group.add_argument(
            "--temporary-allocations",
            help="Equivalent to --temporary-allocation-threshold=1",
            action="store_const",
            dest="temporary_allocation_threshold",
            const=1,
        )
        parser.add_argument("results", help="Results of the tracker run")

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        temporal = getattr(args, "temporal", False)
        if temporal and args.temporary_allocation_threshold >= 0:
            parser.error("Can't create a temporal flame graph of temporary allocations")

        result_path, output_file = self.validate_filenames(
            output=args.output,
            results=args.results,
            overwrite=args.force,
        )
        self.output_file = output_file
        kwargs = {}
        if hasattr(args, "split_threads"):
            kwargs["merge_threads"] = not args.split_threads

        if hasattr(args, "inverted"):
            kwargs["inverted"] = args.inverted

        if hasattr(args, "max_memory_records"):
            kwargs["max_memory_records"] = args.max_memory_records

        if hasattr(args, "no_web"):
            kwargs["no_web"] = args.no_web

        self.write_report(
            result_path,
            output_file,
            args.show_memory_leaks,
            args.temporary_allocation_threshold,
            temporal=temporal,
            **kwargs,
        )

        print(f"Wrote {output_file}")



import argparse
import os

from memray import dump_all_records
from memray._errors import MemrayCommandError


class ParseCommand:
    """Debug a results file by parsing and printing each record in it"""

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("results", help="Results of the tracker run")

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        if os.isatty(1):
            raise MemrayCommandError(
                "You must redirect stdout to a file or shell pipeline.",
                exit_code=1,
            )

        try:
            dump_all_records(args.results)
        except OSError as e:
            raise MemrayCommandError(
                f"Failed to parse allocation records in {args.results}\nReason: {e}",
                exit_code=1,
            )



import argparse
import ast
import contextlib
import os
import pathlib
import runpy
import socket
import subprocess
import sys
import textwrap
from contextlib import closing
from contextlib import suppress
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from memray import Destination
from memray import FileDestination
from memray import FileFormat
from memray import SocketDestination
from memray import Tracker
from memray._errors import MemrayCommandError
from memray.commands.live import LiveCommand


def _get_free_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind(("", 0))
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return int(sock.getsockname()[1])


def _should_modify_sys_path() -> bool:
    isolated_mode = sys.flags.isolated
    safe_path_mode = getattr(sys.flags, "safe_path", False)  # New in Python 3.11
    return not isolated_mode and not safe_path_mode


def _run_tracker(
    destination: Destination,
    args: argparse.Namespace,
    post_run_message: Optional[str] = None,
) -> None:
    try:
        kwargs: Dict[str, Any] = {}
        if args.follow_fork:
            kwargs["follow_fork"] = True
        if args.trace_python_allocators:
            kwargs["trace_python_allocators"] = True
        if args.aggregate:
            kwargs["file_format"] = FileFormat.AGGREGATED_ALLOCATIONS
        tracker = Tracker(destination=destination, native_traces=args.native, **kwargs)
    except OSError as error:
        raise MemrayCommandError(str(error), exit_code=1)

    with tracker:
        pid = os.getpid()
        try:
            if args.run_as_module:
                if _should_modify_sys_path():
                    sys.path[0] = os.getcwd()
                # run_module will replace argv[0] with the script's path
                sys.argv = ["", *args.script_args]
                runpy.run_module(args.script, run_name="__main__", alter_sys=True)
            elif args.run_as_cmd:
                if _should_modify_sys_path():
                    sys.path[0] = ""
                sys.argv = ["-c", *args.script_args]
                exec(args.script, {"__name__": "__main__"})
            else:
                if _should_modify_sys_path():
                    sys.path[0] = str(
                        pathlib.Path(args.script).resolve().parent.absolute()
                    )
                sys.argv = [args.script, *args.script_args]
                runpy.run_path(args.script, run_name="__main__")
        finally:
            if not args.quiet and post_run_message is not None and pid == os.getpid():
                print(post_run_message)


def _child_process(
    port: int,
    native: bool,
    trace_python_allocators: bool,
    run_as_module: bool,
    run_as_cmd: bool,
    quiet: bool,
    script: str,
    script_args: List[str],
) -> None:
    args = argparse.Namespace(
        native=native,
        trace_python_allocators=trace_python_allocators,
        follow_fork=False,
        aggregate=False,
        run_as_module=run_as_module,
        run_as_cmd=run_as_cmd,
        quiet=quiet,
        script=script,
        script_args=script_args,
    )
    _run_tracker(destination=SocketDestination(server_port=port), args=args)


def _run_child_process_and_attach(args: argparse.Namespace) -> None:
    port = args.live_port
    if port is None:
        port = _get_free_port()
    if not 2**16 > port > 0:
        raise MemrayCommandError(f"Invalid port: {port}", exit_code=1)

    arguments = (
        f"{port},{args.native},{args.trace_python_allocators},"
        f"{args.run_as_module},{args.run_as_cmd},{args.quiet},"
        f"{args.script!r},{args.script_args}"
    )
    tracked_app_cmd = [
        sys.executable,
        "-c",
        f"from memray.commands.run import _child_process;_child_process({arguments})",
    ]
    with contextlib.suppress(KeyboardInterrupt):
        with subprocess.Popen(
            tracked_app_cmd,
            stderr=subprocess.PIPE,
            stdout=subprocess.DEVNULL,
            text=True,
        ) as process:
            try:
                LiveCommand().start_live_interface(
                    port, cmdline_override=" ".join(sys.argv)
                )
            except (Exception, KeyboardInterrupt) as error:
                process.terminate()
                raise error from None
            process.terminate()
            if process.returncode:
                if process.stderr:
                    print(process.stderr.read(), file=sys.stderr)
                raise (MemrayCommandError(exit_code=process.returncode))


def _run_with_socket_output(args: argparse.Namespace) -> None:
    port = args.live_port
    if port is None:
        port = _get_free_port()
    if not 2**16 > port > 0:
        raise MemrayCommandError(f"Invalid port: {port}", exit_code=1)

    if not args.quiet:
        memray_cli = f"memray{sys.version_info.major}.{sys.version_info.minor}"
        print(f"Run '{memray_cli} live {port}' in another shell to see live results")
    with suppress(KeyboardInterrupt):
        _run_tracker(destination=SocketDestination(server_port=port), args=args)


def _run_with_file_output(args: argparse.Namespace) -> None:
    if args.output is None:
        script_name = args.script
        if args.run_as_cmd:
            script_name = "string"

        output = f"memray-{os.path.basename(script_name)}.{os.getpid()}.bin"
        filename = os.path.join(os.path.dirname(script_name), output)
    else:
        filename = args.output

    if not args.quiet:
        print(f"Writing profile results into {filename}")

    example_report_generation_message = textwrap.dedent(
        f"""
        [memray] Successfully generated profile results.

        You can now generate reports from the stored allocation records.
        Some example commands to generate reports:

        {sys.executable} -m memray flamegraph {filename}
        """
    ).strip()

    destination = FileDestination(
        path=filename, overwrite=args.force, compress_on_exit=args.compress_on_exit
    )
    try:
        _run_tracker(
            destination=destination,
            args=args,
            post_run_message=example_report_generation_message,
        )
    except OSError as error:
        raise MemrayCommandError(str(error), exit_code=1)


class RunCommand:
    """Run the specified application and track memory usage"""

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.usage = "%(prog)s [-m module | -c cmd | file] [args]"
        output_group = parser.add_mutually_exclusive_group()
        output_group.add_argument(
            "-o",
            "--output",
            help="Output file name (default: <process_name>.<pid>.bin)",
        )
        output_group.add_argument(
            "--live",
            help="Start a live tracking session and immediately connect a live server",
            action="store_true",
            dest="live_mode",
            default=False,
        )
        output_group.add_argument(
            "--live-remote",
            help="Start a live tracking session and wait until a client connects",
            action="store_true",
            dest="live_remote_mode",
            default=False,
        )
        parser.add_argument(
            "--live-port",
            "-p",
            help="Port to use when starting live tracking (default: random free port)",
            default=None,
            type=int,
        )
        parser.add_argument(
            "--aggregate",
            help="Write aggregated stats to the output file instead of all allocations",
            action="store_true",
            default=False,
        )

        parser.add_argument(
            "--native",
            help="Track native (C/C++) stack frames as well",
            action="store_true",
            dest="native",
            default=False,
        )
        parser.add_argument(
            "--follow-fork",
            action="store_true",
            help="Record allocations in child processes forked from the tracked script",
            default=False,
        )
        parser.add_argument(
            "--trace-python-allocators",
            action="store_true",
            help="Record allocations made by the pymalloc allocator",
            default=False,
        )
        parser.add_argument(
            "-q",
            "--quiet",
            help="Don't show any tracking-specific output while running",
            action="store_true",
        )
        parser.add_argument(
            "-f",
            "--force",
            help="If the output file already exists, overwrite it",
            action="store_true",
            default=False,
        )
        compression = parser.add_mutually_exclusive_group()
        compression.add_argument(
            "--compress-on-exit",
            help="Compress the resulting file using lz4 after tracking completes",
            default=True,
            action="store_true",
        )
        compression.add_argument(
            "--no-compress",
            help="Do not compress the resulting file using lz4",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "-c",
            help="Program passed in as string",
            action="store_true",
            dest="run_as_cmd",
            default=False,
        )
        parser.add_argument(
            "-m",
            help="Run library module as a script (terminates option list)",
            action="store_true",
            dest="run_as_module",
        )
        parser.add_argument("script", help=argparse.SUPPRESS, metavar="file")
        parser.add_argument(
            "script_args",
            help=argparse.SUPPRESS,
            nargs=argparse.REMAINDER,
            metavar="module",
        )

    def validate_target_file(self, args: argparse.Namespace) -> None:
        """Ensure we are running a Python file"""
        if args.run_as_module:
            return
        try:
            if args.run_as_cmd:
                source = bytes(args.script, "UTF-8")
            else:
                source = pathlib.Path(args.script).read_bytes()
            ast.parse(source)
        except (SyntaxError, ValueError):
            raise MemrayCommandError(
                "Only valid Python files or commands can be executed under memray",
                exit_code=1,
            )

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        if args.no_compress:
            args.compress_on_exit = False

        if args.live_port is not None and not args.live_remote_mode:
            parser.error("The --live-port argument requires --live-remote")
        if args.follow_fork is True and (args.live_mode or args.live_remote_mode):
            parser.error("--follow-fork cannot be used with the live TUI")
        if args.aggregate and (args.live_mode or args.live_remote_mode):
            parser.error("--aggregate cannot be used with the live TUI")
        with contextlib.suppress(OSError):
            if args.run_as_cmd and pathlib.Path(args.script).exists():
                parser.error("remove the option -c to run a file")

        self.validate_target_file(args)

        if args.live_mode:
            _run_child_process_and_attach(args)
        elif args.live_remote_mode:
            _run_with_socket_output(args)
        else:
            _run_with_file_output(args)



from dataclasses import dataclass

from ._memray import PythonStackElement
from ._metadata import Metadata

@dataclass
class Stats:
    metadata: Metadata
    total_num_allocations: int
    total_memory_allocated: int
    peak_memory_allocated: int
    allocation_count_by_size: dict[int, int]
    allocation_count_by_allocator: dict[str, int]
    top_locations_by_size: list[tuple[PythonStackElement, int]]
    top_locations_by_count: list[tuple[PythonStackElement, int]]



from typing import Callable

from ._memray import PymallocDomain as PymallocDomain

class MemoryAllocator:
    def __init__(self) -> None: ...
    def free(self) -> None: ...
    def malloc(self, size: int) -> bool: ...
    def calloc(self, size: int) -> bool: ...
    def realloc(self, size: int) -> bool: ...
    def posix_memalign(self, size: int) -> bool: ...
    def aligned_alloc(self, size: int) -> bool: ...
    def memalign(self, size: int) -> bool: ...
    def valloc(self, size: int) -> bool: ...
    def pvalloc(self, size: int) -> bool: ...
    def run_in_pthread(self, callback: Callable[[], None]) -> None: ...

class PymallocMemoryAllocator:
    def __init__(self, domain: PymallocDomain) -> None: ...
    def free(self) -> None: ...
    def malloc(self, size: int) -> None: ...
    def calloc(self, size: int) -> None: ...
    def realloc(self, size: int) -> None: ...

class MmapAllocator:
    def __init__(self, size: int, address: int = 0) -> None: ...
    @property
    def address(self) -> int: ...
    def munmap(self, length: int, offset: int = 0) -> None: ...

def _cython_nested_allocation(
    allocator_fn: Callable[[int], None], size: int
) -> None: ...
def _cython_allocate_in_two_places(size: int) -> None: ...
def set_thread_name(name: str) -> int: ...
def function_caller(func: Callable[[], None]) -> None: ...
def allocate_without_gil_held(wake_up_main_fd: int, wake_up_thread_fd: int) -> None: ...
def allocate_cpp_vector(size: int) -> int: ...
def fill_cpp_vector(size: int) -> int: ...
def exit(py_finalize: bool = False) -> None: ...

class PrimeCaches:
    def __init__(self, size: int) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, *args: object) -> None: ...



import argparse
import logging
import sys
import textwrap
from typing import List
from typing import Optional

from memray._version import __version__

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol

from memray._errors import MemrayCommandError
from memray._errors import MemrayError
from memray._memray import set_log_level

from . import attach
from . import flamegraph
from . import live
from . import parse
from . import run
from . import stats
from . import summary
from . import table
from . import transform
from . import tree

_EPILOG = textwrap.dedent(
    """\
    Please submit feedback, ideas, and bug reports by filing a new issue at
    https://github.com/bloomberg/memray/issues
    """
)

_DESCRIPTION = textwrap.dedent(
    """\
    Memory profiler for Python applications

    Run `memray run` to generate a memory profile report, then use a reporter command
    such as `memray flamegraph` or `memray table` to convert the results into HTML.

    Example:

        $ python3 -m memray run -o output.bin my_script.py
        $ python3 -m memray flamegraph output.bin
    """
)


class Command(Protocol):
    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        ...

    def run(self, args: argparse.Namespace, parser: argparse.ArgumentParser) -> None:
        ...


_COMMANDS: List[Command] = [
    run.RunCommand(),
    flamegraph.FlamegraphCommand(),
    table.TableCommand(),
    live.LiveCommand(),
    tree.TreeCommand(),
    parse.ParseCommand(),
    summary.SummaryCommand(),
    stats.StatsCommand(),
    transform.TransformCommand(),
    attach.AttachCommand(),
    attach.DetachCommand(),
]


def get_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=_DESCRIPTION,
        prog="memray",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=_EPILOG,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity. Option is additive and can be specified up to 3 times",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=__version__,
        help="Displays the current version of Memray",
    )

    subparsers = parser.add_subparsers(
        help="Mode of operation",
        dest="command",
        required=True,
    )

    for command in _COMMANDS:
        # Extract the CLI command name from the classes' names
        assert command.__class__.__name__.endswith("Command")
        name = command.__class__.__name__[: -len("Command")].lower()

        # Add the subcommand
        command_parser = subparsers.add_parser(
            name, help=command.__doc__, description=command.__doc__, epilog=_EPILOG
        )
        command_parser.set_defaults(entrypoint=command.run)
        command.prepare_parser(command_parser)

    return parser


def determine_logging_level_from_verbosity(
    verbose_level: int,
) -> int:  # pragma: no cover
    if verbose_level == 0:
        return logging.WARNING
    elif verbose_level == 1:
        return logging.INFO
    else:
        return logging.DEBUG


def main(args: Optional[List[str]] = None) -> int:
    if args is None:
        args = sys.argv[1:]

    parser = get_argument_parser()
    arg_values = parser.parse_args(args=args)
    set_log_level(determine_logging_level_from_verbosity(arg_values.verbose))

    try:
        arg_values.entrypoint(arg_values, parser)
    except MemrayCommandError as e:
        print(e, file=sys.stderr)
        return e.exit_code
    except MemrayError as e:
        print(e, file=sys.stderr)
        return 1
    else:
        return 0



import os
from typing import IO
from typing import Iterable
from typing import Optional

from rich import print as rprint
from rich.markup import escape
from rich.table import Column
from rich.table import Table

from memray import AllocationRecord
from memray._memray import size_fmt
from memray.reporters.tui import aggregate_allocations

MAX_MEMORY_RATIO = 0.95
DEFAULT_TERMINAL_LINES = 24


def _get_terminal_lines() -> int:
    try:
        return os.get_terminal_size().lines
    except OSError:
        return DEFAULT_TERMINAL_LINES


def _size_to_color(proportion_of_total: float) -> str:
    if proportion_of_total > 0.6:
        return "red"
    elif proportion_of_total > 0.2:
        return "yellow"
    elif proportion_of_total > 0.05:
        return "green"
    else:
        return "bright_green"


class SummaryReporter:
    KEY_TO_COLUMN_NAME = {
        1: "total_memory",
        2: "total_memory",
        3: "own_memory",
        4: "own_memory",
        5: "n_allocations",
    }

    N_COLUMNS = len(KEY_TO_COLUMN_NAME)

    def __init__(self, data: Iterable[AllocationRecord], native: bool):
        snapshot = tuple(data)
        self.current_memory_size = sum(record.size for record in snapshot)
        self.total_allocations = sum(record.n_allocations for record in snapshot)
        self.snapshot_data = aggregate_allocations(
            snapshot,
            MAX_MEMORY_RATIO * self.current_memory_size,
            native,
        )

    @classmethod
    def from_snapshot(
        cls, allocations: Iterable[AllocationRecord], native: bool = False
    ) -> "SummaryReporter":
        return cls(allocations, native=native)

    def render(
        self,
        sort_column: int,
        *,
        max_rows: Optional[int] = None,
        file: Optional[IO[str]] = None,
    ) -> None:
        # If no rows need to be split across 2 lines, using 5 fewer rows in
        # the data table than the number of terminal lines lets us fit entirely
        # on 1 full page regardless of whether the user has a 1 line or 2 line
        # shell prompt. With a 2 line prompt we lose the top dashed line.
        max_rows = max_rows or max(_get_terminal_lines() - 5, 10)
        table = Table(
            Column("Location", ratio=5),
            Column("Total Memory", ratio=1, justify="right"),
            Column("Total Memory %", ratio=1, justify="right"),
            Column("Own Memory", ratio=1, justify="right"),
            Column("Own Memory % ", ratio=1, justify="right"),
            Column("Allocation Count", ratio=1, justify="right"),
            expand=True,
        )
        table.columns[sort_column].header = f"<{table.columns[sort_column].header}>"

        sorted_allocations = sorted(
            self.snapshot_data.items(),
            key=lambda item: getattr(item[1], self.KEY_TO_COLUMN_NAME[sort_column]),
            reverse=True,
        )[:max_rows]
        for location, result in sorted_allocations:
            color_location = (
                f"[bold magenta]{escape(location.function)}[/] at "
                f"[cyan]{escape(location.file)}[/]"
            )
            total_color = _size_to_color(result.total_memory / self.current_memory_size)
            own_color = _size_to_color(result.own_memory / self.current_memory_size)
            allocation_colors = _size_to_color(
                result.n_allocations / self.total_allocations
            )
            percent_total = result.total_memory / self.current_memory_size * 100
            percent_own = result.own_memory / self.current_memory_size * 100
            table.add_row(
                color_location,
                f"[{total_color}]{size_fmt(result.total_memory)}[/{total_color}]",
                f"[{total_color}]{percent_total:.2f}%[/{total_color}]",
                f"[{own_color}]{size_fmt(result.own_memory)}[/{own_color}]",
                f"[{own_color}]{percent_own:.2f}%[/{own_color}]",
                f"[{allocation_colors}]{result.n_allocations}[/{allocation_colors}]",
            )

        rprint(table, file=file)
